/*****************************************/
/*                                       */
/*         MBS270 demo server            */
/*                                       */
/* Copyright Mobisense Systems 2007      */
/*                                       */
/* Gradient computations code.           */
/*                                       */
/*****************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>

#include "ippdefs.h"
#include "ippIP.h"

#include "gradient_computations.h"

/* #define GRADIENT_COMPUTATIONS_DEBUG */


int mds_init_horizontal_gradient(int width, int height, mds_grad_attr_t *grad)
{
  int ii;
  int border_width = GRAD_KERNEL_SIZE>>1;

  /* for column filter, borders are only top and bottom lines */
  grad->imgOffset = border_width * width;
  grad->imgStep   = width; 

  /* column filter attributes */
  /* the result buffer does not include top and bottom border lines */
  grad->flt_roiSize.width  = width;
  grad->flt_roiSize.height = height - 2 * border_width;
  grad->fltBuf = malloc(grad->flt_roiSize.width * grad->flt_roiSize.height);
  if (grad->fltBuf == NULL) {
    perror("malloc");
    return -1;
  }
  grad->fltStep = width;
  grad->kernel[0] = 1;
  grad->kernel[1] = 2;
  grad->kernel[2] = 1;
  grad->kernelSize = GRAD_KERNEL_SIZE;
  grad->kernelAnchor = 1;
  grad->divider = 0;
  for (ii=0; ii<GRAD_KERNEL_SIZE; ii++) grad->divider += grad->kernel[ii];

  /* Result buffer attributes */
  /* this buffer contains gradient image and computation time at the end */
  grad->timeOffset = width * height;
  grad->bufSize = grad->timeOffset + sizeof(int);
  grad->resBuf  = malloc(grad->bufSize);
  if (grad->resBuf == NULL) {
    perror("malloc");
    return -1;
  }
  grad->resOffset = width + border_width;
  grad->resStep   = width;
  grad->sub_roiSize.width  = width - 2 * border_width;
  grad->sub_roiSize.height = grad->flt_roiSize.height;
  /* initialize borders with null value */
  IppiSize roiSize;
  roiSize.width = width;
  roiSize.height = height;
  ippiSet_8u_C1R(GRAD_NULL_VALUE, grad->resBuf, grad->resStep, roiSize);

#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("mds_init_horizontal_gradient\n");
  printf("fltBuf: roi.w = %d, roi.h = %d, imgOffset = %d, kernel = {%d,%d,%d}, divider = %d\n",
	 grad->flt_roiSize.width,
	 grad->flt_roiSize.height,
	 grad->imgOffset,
	 grad->kernel[0], grad->kernel[1], grad->kernel[2], grad->divider);
  printf("resBuf: roi.w = %d, roi.h = %d, resOffset = %d\n",
	 grad->sub_roiSize.width,
	 grad->sub_roiSize.height,
	 grad->resOffset);
#endif
  return 0;
}


void mds_deinit_horizontal_gradient(mds_grad_attr_t *grad)
{
  if (grad->fltBuf != NULL) { free(grad->fltBuf); grad->fltBuf = NULL; }
  if (grad->resBuf != NULL) { free(grad->resBuf); grad->resBuf = NULL; }
}


int mds_compute_horizontal_gradient(unsigned char *img, mds_grad_attr_t *grad)
{
  int ii,jj;
  struct timeval tvs, tve;
  gettimeofday(&tvs, NULL);
  
  IppStatus status;
  status = ippiFilterColumn_8u_C1R(img + grad->imgOffset,
				   grad->imgStep,
				   grad->fltBuf,
				   grad->fltStep,
				   grad->flt_roiSize,
				   grad->kernel,
				   grad->kernelSize,
				   grad->kernelAnchor,
				   grad->divider);
  if (status != ippStsNoErr) {
    printf("ippiFilterColumn_8u_C1R ERROR, status = %d\n", status);
    return -1;
  }

  /* substract with a loop to avoid saturation */
  char *res = grad->resBuf + grad->resOffset;
  char *flt = grad->fltBuf;
  int val;
  for (ii=0; ii<grad->sub_roiSize.height; ii++) {
    for (jj=0; jj<grad->sub_roiSize.width; jj++) {
      *(res+jj) = 128 + ((*(flt+jj+2) - *(flt+jj)) >> 1);
    }
    res += grad->resStep;
    flt += grad->fltStep;
  }

  gettimeofday(&tve, NULL);
  int dt = tve.tv_usec - tvs.tv_usec;
  if (dt < 0) dt += 1000000;
#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("dt = %dus\n", dt);
#endif
  *((int*) (grad->resBuf + grad->timeOffset)) = dt;

  return 0;
}



int mds_init_vertical_gradient(int width, int height, mds_grad_attr_t *grad)
{
  int ii;
  int border_width = GRAD_KERNEL_SIZE>>1;

  /* for raw filter, borders are only left and right columns */
  grad->imgOffset = border_width;
  grad->imgStep   = width; 

  /* raw filter attributes */
  /* the result buffer does not include left and right border columns */
  grad->flt_roiSize.width  = width - 2 * border_width;
  grad->flt_roiSize.height = height;
  grad->fltBuf = malloc(grad->flt_roiSize.width * grad->flt_roiSize.height);
  if (grad->fltBuf == NULL) {
    perror("malloc");
    return -1;
  }
  grad->fltStep = grad->flt_roiSize.width;
  grad->kernel[0] = 1;
  grad->kernel[1] = 2;
  grad->kernel[2] = 1;
  grad->kernelSize = GRAD_KERNEL_SIZE;
  grad->kernelAnchor = 1;
  grad->divider = 0;
  for (ii=0; ii<GRAD_KERNEL_SIZE; ii++) grad->divider += grad->kernel[ii];

  /* Result buffer attributes */
  /* this buffer contains gradient image and computation time at the end */
  grad->timeOffset = width * height;
  grad->bufSize = grad->timeOffset + sizeof(int);
  grad->resBuf  = malloc(grad->bufSize);
  if (grad->resBuf == NULL) {
    perror("malloc");
    return -1;
  }
  grad->resOffset = width + border_width;
  grad->resStep   = width;
  grad->sub_roiSize.width  = grad->flt_roiSize.width;
  grad->sub_roiSize.height = height - 2 * border_width;
  /* initialize borders with null value */
  IppiSize roiSize;
  roiSize.width = width;
  roiSize.height = height;
  ippiSet_8u_C1R(GRAD_NULL_VALUE, grad->resBuf, grad->resStep, roiSize);

#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("mds_init_vertical_gradient\n");
  printf("fltBuf: roi.w = %d, roi.h = %d, imgOffset = %d, kernel = {%d,%d,%d}, divider = %d\n",
	 grad->flt_roiSize.width,
	 grad->flt_roiSize.height,
	 grad->imgOffset,
	 grad->kernel[0], grad->kernel[1], grad->kernel[2], grad->divider);
  printf("resBuf: roi.w = %d, roi.h = %d, resOffset = %d\n",
	 grad->sub_roiSize.width,
	 grad->sub_roiSize.height,
	 grad->resOffset);
#endif
  return 0;
}


void mds_deinit_vertical_gradient(mds_grad_attr_t *grad)
{
  if (grad->fltBuf != NULL) { free(grad->fltBuf); grad->fltBuf = NULL; }
  if (grad->resBuf != NULL) { free(grad->resBuf); grad->resBuf = NULL; }
}


int mds_compute_vertical_gradient(unsigned char *img, mds_grad_attr_t *grad)
{
  int ii,jj;
  struct timeval tvs, tve;
  gettimeofday(&tvs, NULL);
  
  IppStatus status;
  status = ippiFilterRow_8u_C1R(img + grad->imgOffset,
				grad->imgStep,
				grad->fltBuf,
				grad->fltStep,
				grad->flt_roiSize,
				grad->kernel,
				grad->kernelSize,
				grad->kernelAnchor,
				grad->divider);
  if (status != ippStsNoErr) {
    printf("ippiFilterRow_8u_C1R ERROR, status = %d\n", status);
    return -1;
  }

  /* substract with a loop to avoid saturation */
  char *res = grad->resBuf + grad->resOffset;
  char *flt_n = grad->fltBuf + 2 * grad->fltStep;
  char *flt_p = grad->fltBuf;
  for (ii=0; ii<grad->sub_roiSize.height; ii++) {
    for (jj=0; jj<grad->sub_roiSize.width; jj++) {
      *(res+jj) = 128 + ((*(flt_n+jj) - *(flt_p+jj)) >> 1);
    }
    res += grad->resStep;
    flt_n += grad->fltStep;
    flt_p += grad->fltStep;
  }

  gettimeofday(&tve, NULL);
  int dt = tve.tv_usec - tvs.tv_usec;
  if (dt < 0) dt += 1000000;
#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("dt = %dus\n", dt);
#endif
  *((int*) (grad->resBuf + grad->timeOffset)) = dt;

  return 0;
}


int mds_init_edge(int width, int height, mds_edge_attr_t *edge)
{
  int ii;
  int border_width = EDGE_KERNEL_SIZE>>1;

  edge->imgStep   = width; 

  /* for column filter, borders are only top and bottom lines */
  edge->imghOffset = border_width * width;

  /* column filter attributes */
  /* the result buffer does not include top and bottom border lines */
  edge->flth_roiSize.width  = width;
  edge->flth_roiSize.height = height - 2 * border_width;
  edge->flthBuf = malloc(edge->flth_roiSize.width * edge->flth_roiSize.height);
  if (edge->flthBuf == NULL) {
    perror("malloc");
    return -1;
  }
  edge->flthStep = width;
  edge->kernelh[0] = 1;
  edge->kernelh[1] = 2;
  edge->kernelh[2] = 1;
  edge->kernelhSize = EDGE_KERNEL_SIZE;
  edge->kernelhAnchor = 1;
  edge->dividerh = 0;
  for (ii=0; ii<EDGE_KERNEL_SIZE; ii++) edge->dividerh += edge->kernelh[ii];

  /* for raw filter, borders are only left and right columns */
  edge->imgvOffset = border_width;

  /* raw filter attributes */
  /* the result buffer does not include left and right border columns */
  edge->fltv_roiSize.width  = width - 2 * border_width;
  edge->fltv_roiSize.height = height;
  edge->fltvBuf = malloc(edge->fltv_roiSize.width * edge->fltv_roiSize.height);
  if (edge->fltvBuf == NULL) {
    perror("malloc");
    return -1;
  }
  edge->fltvStep = edge->fltv_roiSize.width;
  edge->kernelv[0] = 1;
  edge->kernelv[1] = 2;
  edge->kernelv[2] = 1;
  edge->kernelvSize = EDGE_KERNEL_SIZE;
  edge->kernelvAnchor = 1;
  edge->dividerv = 0;
  for (ii=0; ii<EDGE_KERNEL_SIZE; ii++) edge->dividerv += edge->kernelv[ii];

  /* Result buffer attributes */
  /* this buffer contains edge image and computation time at the end */
  edge->timeOffset = width * height;
  edge->bufSize = edge->timeOffset * sizeof(*edge->resBuf) + sizeof(int);
  edge->resBuf  = calloc(1, edge->bufSize); /* set borders to zero */
  if (edge->resBuf == NULL) {
    perror("calloc");
    return -1;
  }
  edge->resOffset = width + border_width;
  edge->resStep   = width;
  edge->sub_roiSize.width  = edge->fltv_roiSize.width;
  edge->sub_roiSize.height = edge->flth_roiSize.height;

#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("mds_init_edge\n");
  printf("flthBuf: roi.w = %d, roi.h = %d, imgOffset = %d, kernel = {%d,%d,%d}, divider = %d\n",
	 edge->flth_roiSize.width,
	 edge->flth_roiSize.height,
	 edge->imghOffset,
	 edge->kernelh[0], edge->kernelh[1], edge->kernelh[2], edge->dividerh);
  printf("fltvBuf: roi.w = %d, roi.h = %d, imgOffset = %d, kernel = {%d,%d,%d}, divider = %d\n",
	 edge->fltv_roiSize.width,
	 edge->fltv_roiSize.height,
	 edge->imgvOffset,
	 edge->kernelv[0], edge->kernelv[1], edge->kernelv[2], edge->dividerv);
  printf("resBuf: roi.w = %d, roi.h = %d, resOffset = %d\n",
	 edge->sub_roiSize.width,
	 edge->sub_roiSize.height,
	 edge->resOffset);
  printf("bufSize = %d, timeOffset = %d\n", edge->bufSize, edge->timeOffset);
#endif
  return 0;
}


void mds_deinit_edge(mds_edge_attr_t *edge)
{
  if (edge->flthBuf != NULL) { free(edge->flthBuf); edge->flthBuf = NULL; }
  if (edge->fltvBuf != NULL) { free(edge->fltvBuf); edge->fltvBuf = NULL; }
  if (edge->resBuf  != NULL) { free(edge->resBuf ); edge->resBuf  = NULL; }
}


int mds_compute_edge(unsigned char *img, mds_edge_attr_t *edge)
{
  int ii,jj;
  struct timeval tvs, tve;
  gettimeofday(&tvs, NULL);
  
  IppStatus status;
  status = ippiFilterColumn_8u_C1R(img + edge->imghOffset,
				   edge->imgStep,
				   edge->flthBuf,
				   edge->flthStep,
				   edge->flth_roiSize,
				   edge->kernelh,
				   edge->kernelhSize,
				   edge->kernelhAnchor,
				   edge->dividerh);
  if (status != ippStsNoErr) {
    printf("ippiFilterColumn_8u_C1R ERROR, status = %d\n", status);
    return -1;
  }
  
  status = ippiFilterRow_8u_C1R(img + edge->imgvOffset,
				edge->imgStep,
				edge->fltvBuf,
				edge->fltvStep,
				edge->fltv_roiSize,
				edge->kernelv,
				edge->kernelvSize,
				edge->kernelvAnchor,
				edge->dividerv);
  if (status != ippStsNoErr) {
    printf("ippiFilterRow_8u_C1R ERROR, status = %d\n", status);
    return -1;
  }

  /* compute with a loop to avoid saturation */
  unsigned short *res = edge->resBuf + edge->resOffset;
  char *flt = edge->flthBuf;
  char *flt_n = edge->fltvBuf + 2 * edge->fltvStep;
  char *flt_p = edge->fltvBuf;
  unsigned short diff;
  for (ii=0; ii<edge->sub_roiSize.height; ii++) {
    for (jj=0; jj<edge->sub_roiSize.width; jj++) {
      diff = (*(flt+jj+2) - *(flt+jj)) >> 1; /* horizontal gradient */
      *(res+jj) = diff * diff;
      diff = (*(flt_n+jj) - *(flt_p+jj)) >> 1; /* vertical gradient */
      *(res+jj) += diff * diff;
    }
    res += edge->resStep;
    flt += edge->flthStep;
    flt_n += edge->fltvStep;
    flt_p += edge->fltvStep;
  }

  gettimeofday(&tve, NULL);
  int dt = tve.tv_usec - tvs.tv_usec;
  if (dt < 0) dt += 1000000;
#ifdef GRADIENT_COMPUTATIONS_DEBUG
  printf("dt = %dus\n", dt);
#endif
  *((int*) (edge->resBuf + edge->timeOffset)) = dt;

  return 0;
}
