#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/fb.h>

#define FB_DEV_NAME "/dev/fb0"

struct frame_buffer {
  void *start;
  size_t length;
};

struct fb_context {
  int   fd;
  struct frame_buffer buf;
};


int fb_init(struct fb_context *ctx)
{
  int ret;
  struct fb_fix_screeninfo   fb_fix;
  struct fb_var_screeninfo   fb_var;

  ctx->fd = -1;

  ctx->fd = open(FB_DEV_NAME, O_RDWR);
  if (ctx->fd < 0) {
    printf("%s ", FB_DEV_NAME);
    perror("open failed");
    return -1;
  }

  printf("%s open succeeded.\n", FB_DEV_NAME);

  /* Turn LCD Off */
  ret = ioctl(ctx->fd, FBIOBLANK, 1);
  if (ret < 0) {
    perror("ioctl FBIOBLANK");
    return ret;
  }

  ret = ioctl(ctx->fd, FBIOGET_FSCREENINFO, &fb_fix);
  if (ret < 0) {
    perror("ioctl FBIOGET_FSCREENINFO");
    return ret;
  }

  printf("id         : %s\n",     fb_fix.id);
  printf("smem_start : 0x%08x\n", fb_fix.smem_start);
  printf("smem_len   : %u\n",     fb_fix.smem_len);
  printf("type       : %u\n",     fb_fix.type);
  printf("visual     : %u\n",     fb_fix.visual);
  printf("line_length: %u\n",     fb_fix.line_length);
  printf("mmio_start : 0x%08x\n", fb_fix.mmio_start);
  printf("mmio_len   : %u\n",     fb_fix.mmio_len);

  ret = ioctl(ctx->fd, FBIOGET_VSCREENINFO, &fb_var);
  if (ret < 0) {
    perror("ioctl FBIOGET_VSCREENINFO");
    return ret;
  }

  printf("xres       : %u\n",     fb_var.xres);
  printf("yres       : %u\n",     fb_var.yres);
  printf("xoffset    : %u\n",     fb_var.xoffset);
  printf("yoffset    : %u\n",     fb_var.yoffset);
  printf("bits_per_p : %u\n",     fb_var.bits_per_pixel);
  printf("grayscale  : %u\n",     fb_var.grayscale);
  printf("activate   : %u\n",     fb_var.activate);

  ctx->buf.length = fb_fix.smem_len;
  ctx->buf.start = mmap (NULL, fb_fix.smem_len,
			 PROT_READ | PROT_WRITE, MAP_SHARED,
			 ctx->fd, 0);
  if (MAP_FAILED == ctx->buf.start) {
    perror ("mmap");
    return -1;
  }

  printf("buf.start  : 0x%08x\n", ctx->buf.start);
  printf("buf.length : %u\n", ctx->buf.length);


  return 0;
}

int fb_display(struct fb_context *ctx)
{
  int ret;
  int ii, *pbuf;
  unsigned char *p18bpp;

  /* Turn LCD On */
  ret = ioctl(ctx->fd, FBIOBLANK, 0);
  if (ret < 0) {
    perror("ioctl FBIOBLANK");
    return ret;
  }

  pbuf = ctx->buf.start;
  ii = 0;
  while ((void*)pbuf < (ctx->buf.start + ctx->buf.length - sizeof(int))) {
    *pbuf = ii;
    ii++;
    pbuf++;
  }
  sleep(2);

  /* Blue screen */
  p18bpp = ctx->buf.start;
  while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length-3)) {
    *p18bpp = 0x3f;
    p18bpp++;
    *p18bpp = 0;
    p18bpp++;
    *p18bpp = 0;
    p18bpp++;
  }
  sleep(2);

  /* Green screen */
  p18bpp = ctx->buf.start;
/*   while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length*2/3-3)) { */
  while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length-3)) {
    *p18bpp = 0xc0;
    p18bpp++;
    *p18bpp = 0x0f;
    p18bpp++;
    *p18bpp = 0;
    p18bpp++;
  }
  sleep(2);

  /* Red screen */
  p18bpp = ctx->buf.start;
/*   while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length/3-3)) { */
  while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length-3)) {
    *p18bpp = 0xc0;
    p18bpp++;
    *p18bpp = 0xf0;
    p18bpp++;
    *p18bpp = 0x03;
    p18bpp++;
  }
  sleep(2);

  /* All colors */
  for (ii=0; ii<(1<<18); ii++) {
    p18bpp = ctx->buf.start;
    while ((void*)p18bpp < (ctx->buf.start + ctx->buf.length-3)) {
      *p18bpp = *((unsigned char *) &ii + 0);
      p18bpp++;
      *p18bpp = *((unsigned char *) &ii + 1);
      p18bpp++;
      *p18bpp = *((unsigned char *) &ii + 2);
      p18bpp++;
    }
  }

  return 0;
}

void fb_end(struct fb_context *ctx)
{
  if ((ctx->buf.start != MAP_FAILED) && (ctx->buf.length > 0))
    munmap (ctx->buf.start, ctx->buf.length);

  /* Turn LCD Off */
  ioctl(ctx->fd, FBIOBLANK, 1);

  if (ctx->fd != -1) close(ctx->fd);
}


int main(int argc, char *argv[])
{
  int ret;
  struct fb_context ctx;
  
  printf("\n\n   ***   test_fb   ***\n\n");
  printf("Basic frame buffer test\n\n");
  usleep(100000); /* for proper printing */

  ret = fb_init(&ctx);
  if (ret < 0) goto end;

  ret = fb_display(&ctx);
  if (ret < 0) goto end;

 end:
  fb_end(&ctx);

  exit(0);
}
