#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <linux/ioctl.h>
#include <linux/pxa_gpio.h>
#include <linux/pxa_timer.h>

#define GPIO_DEV_NAME "/dev/pxa-gpio"
#define TIMER_DEV_NAME "/dev/pxa-timer"

static int stop_timer_test = 0;


void sigint_handler(int sig) {
  if (sig == SIGINT) stop_timer_test = 1;
}


void *thread_check_output(void *arg)
{
  int gpio_dev;
  short gpio_state;
  struct pxa_gpio_add gpio_add;
  struct pxa_gpio_get gpio_get;
  struct timeval tv;

  printf("thread_check_output starting...\n");
  
  gpio_dev = open(GPIO_DEV_NAME, O_RDWR);
  if (gpio_dev < 0) {
    printf("%s ", GPIO_DEV_NAME);
    perror("open failed");
    pthread_exit((void*)1);
  }
  gpio_add.gpio  = *((unsigned int *) arg);
  gpio_add.input = 1;
  if (ioctl(gpio_dev, PXAIOCGPIOADD, (void*) &gpio_add)) {
    perror("ioctl PXAIOCGPIOADD: ");
  }
  else {
    printf("GPIO %u initial input state = ", gpio_add.gpio);
    gpio_state = gpio_add.init_state;
    if (gpio_state) printf("1\n");
    else printf("0\n");
    gpio_get.gpio = gpio_add.gpio;
    while (!stop_timer_test) {
      if (ioctl(gpio_dev, PXAIOCGPIOGET, (void*) &gpio_get)) {
	perror("ioctl PXAIOCGPIOGET: ");
	break;
      }
      else {
	if ((gpio_get.state) && !(gpio_state)) {
	  gpio_state = gpio_get.state;
	  gettimeofday(&tv, NULL);
	  printf("%ld.%06d GPIO %d set.\n",
		 tv.tv_sec, tv.tv_usec, gpio_add.gpio);
	}
	else if (!(gpio_get.state) && (gpio_state)) {
	  gpio_state = gpio_get.state;
	  gettimeofday(&tv, NULL);
	  printf("%ld.%06d GPIO %d cleared.\n",
		 tv.tv_sec, tv.tv_usec, gpio_add.gpio);
	}
      }
      pthread_yield();
    }
  }

  if (ioctl(gpio_dev, PXAIOCGPIOREM, (void*) &gpio_add.gpio)) {
    perror("ioctl PXAIOCGPIOREM: ");
  }

  close(gpio_dev);
  printf("thread_check_output END.\n");
  pthread_exit(0);
}

int main(int argc, char *argv[])
{
  int timer_dev, ii;
  struct pxa_timer_add_chout timer_add;
  unsigned int gpio;
  pthread_t thread_id;
  void *retval;
  struct sigaction sa, old_sa;

  printf("\n\n   ***   test_timer_lf   ***\n\n");
  printf("Low frequency test for PXA TIMER driver\n\n");
  usleep(100000); /* for proper printing */

  if (argc != 5) goto call_error;
  timer_add.nr = (unsigned int) atoi(argv[1]);
  if (timer_add.nr >= NR_PXA_CHOUT) goto call_error;
  timer_add.f_hz  = (unsigned int) atoi(argv[2]);
  timer_add.ratio = (unsigned int) atoi(argv[3]);
  if (timer_add.ratio < 1) goto call_error;
  gpio = (unsigned int) atoi(argv[4]);

  if (pthread_create(&thread_id, NULL, thread_check_output, (void*) &gpio)) {
    perror("pthread_create");
    exit(1);
  }
  
  timer_dev = open(TIMER_DEV_NAME, O_RDWR);
  if (timer_dev < 0) {
    printf("%s\n", TIMER_DEV_NAME);
    perror("open failed");
    exit(1);
  }

  printf("%s open succeeded.\n", TIMER_DEV_NAME);
  
  /* handling of SIGINT for clean stop */
  sa.sa_handler = sigint_handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(SIGINT, &sa, &old_sa);

  if (ioctl(timer_dev, PXAIOCTMR_ADDCO, (void*) &timer_add)) {
    perror("ioctl PXAIOCTMR_ADDCO: ");
  }
  else {
    float freq = ((float) timer_add.f_hz) / ((float) timer_add.ratio);
    printf("TIMER%d outputs signal at %fHz. CTRL-C to stop the test.\n",
	   timer_add.nr, freq);
    printf("Check GPIO %d toggles every %f second(s).\n", gpio, 1/(2*freq));
    pthread_join(thread_id, &retval);
    if (ioctl(timer_dev, PXAIOCTMR_REMCO, (void*) &timer_add.nr)) {
      perror("ioctl PXAIOCTMR_REMCO: ");
    }
  }

  printf("\n\n   ***   End of test_timer_lf   ***\n\n");
  close(timer_dev);
  sigaction(SIGINT, &old_sa, NULL);
  exit(0);
    
 call_error:
  printf("Error. Call syntax is: test_timer_lf TIMER_number frequency ratio GPIO\n");
  printf("   TIMER_number is the TIMER output number between 0 and %d\n",
	 NR_PXA_CHOUT-1);
  printf("   frequency is the TIMER frequency in Hertz\n");
  printf("   ratio divides frequency to get sub Hertz operation\n");
  printf("   GPIO is the input used to check transitions of timer output\n");
  exit(1);
}
