#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/videodev2.h>
#include <linux/ov7720.h>

#define VIDEO_DEV_NAME "/dev/video0"
#define IMAGE_DIR "/var/tmp"
#define IMAGE_Y_NAME "image_y.pgm"
#define IMAGE_U_NAME "image_u.pgm"
#define IMAGE_V_NAME "image_v.pgm"
#define IMAGE_G_NAME "image_g.pgm"
#define IMAGE_R_NAME "image_r.pgm"
#define IMAGE_B_NAME "image_b.pgm"
#define IMAGE_NAME   "image.ppm"
#define IMAGE_BAYER_NAME   "image_bayer.ppm"

#define OV7720_I2C_ADDR 	0x21
#define OV7720_AGC		0x00
#define OV7720_AECH		0x08
#define OV7720_COM2		0x09
#define OV7720_COM3		0x0C
#define OV7720_AEC		0x10
#define OV7720_CLKRC		0x11
#define OV7720_COM8		0x13
#define OV7720_COM9		0x14
#define OV7720_HSTART		0x0017
#define OV7720_HSIZE		0x0018
#define OV7720_VSTART		0x0019
#define OV7720_VSIZE		0x001A
#define OV7720_LUMHTH		0x0030
#define OV7720_LUMLTH		0x0031
#define OV7720_HREF		0x0032

#define YUYV_STR  "YUYV"
#define UYVY_STR  "UYVY"
#define YUV422P_STR "YUV422P"
#define GRB422_STR  "GRB422"
#define GRB422P_STR "GRB422P"

#define BGGR_2_RGB_ALGO_NNR 0 /* nearest neighbour replication, OK at low res */
#define BGGR_2_RGB_ALGO_AVR 1 /* average, better but more complex */

/* A small function to align a pointer to the (next) page boundary */
/* runs only on 32-bits machines                                   */
unsigned int page_align(unsigned int ptr)
{
  unsigned int page_size;
  page_size = getpagesize();
  return ((ptr + page_size - 1) & ~(page_size - 1));
}


struct image_buffer {
  void *start;
  size_t length;
};

struct grab_context {
  __u32 width;
  __u32 height;
  __u32 pixelformat;
  __u32 i2c_addr;
  int   vfd;
  int   nr_buffers;
  struct image_buffer *buffers;
};

int check_args(int argc, char *argv[], struct grab_context *gctx)
{
  if (argc != 4)
    goto call_error;

  if (strcmp(argv[1],"vga") == 0 || strcmp(argv[1],"VGA") == 0) {
    gctx->width = 640;
    gctx->height = 480;
  }
  else if (strcmp(argv[1],"qvga") == 0 || strcmp(argv[1],"QVGA") == 0) {
    gctx->width = 320;
    gctx->height = 240;
  }
  else
    goto call_error;
  
  if (strcmp(argv[2], UYVY_STR) == 0)
    gctx->pixelformat = V4L2_PIX_FMT_UYVY;
  else if (strcmp(argv[2], YUYV_STR) == 0)
    gctx->pixelformat = V4L2_PIX_FMT_YUYV;
  else if (strcmp(argv[2], YUV422P_STR) == 0)
    gctx->pixelformat = V4L2_PIX_FMT_YUV422P;
  else if (strcmp(argv[2], GRB422_STR) == 0)
    gctx->pixelformat = V4L2_PIX_FMT_SBGGR8;
  else if (strcmp(argv[2], GRB422P_STR) == 0)
    gctx->pixelformat = V4L2_PIX_FMT_GRB422P;
  else
    goto call_error;

  gctx->nr_buffers = atoi(argv[3]);
  if (gctx->nr_buffers <= 0)
    goto call_error;

  return 0;

 call_error:
  printf("Error. Call syntax is: test_grab image_format pixel_format nr_grab\n");
  printf("  image_format: ""vga"" for 640x480 image, ""qvga"" for 320x240.\n");
  printf("  pixel_format: one of the following strings\n");
  printf("                ""%s"" packed UYVY format.\n", UYVY_STR);
  printf("                ""%s"" packed YUYV format.\n", YUYV_STR);
  printf("                ""%s"" planarized YUV422 format.\n", YUV422P_STR);
  printf("                ""%s"" packed GRB422 format.\n", GRB422_STR);
  printf("                ""%s"" planarized GRB422 format.\n", GRB422P_STR);
  printf("  nr_grab     : number of grabbed images.\n");
  return -1;
}


/* set_reg is for DEBUG ONLY */
int set_reg(struct grab_context *gctx,
	    char *reg_name,
	    struct v4l2_register *reg,
	    __u64 *mask)
{
  int ret;
  __u64 new_val;

  new_val = reg->val;

  reg->match_type = V4L2_CHIP_MATCH_I2C_ADDR;
  reg->match_chip = gctx->i2c_addr;
  ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, reg);
  if (ret < 0) {
    perror("ioctl VIDIOC_DBG_G_REGISTER");
    return ret;
  }
  printf("%s (0x%x) current value = 0x%02x\n",
	 reg_name, (int) reg->reg, (int) reg->val);

  reg->val &= ~(*mask);
  reg->val |= new_val;
  ret = ioctl(gctx->vfd, VIDIOC_DBG_S_REGISTER, reg);
  if (ret < 0) {
    perror("ioctl VIDIOC_DBG_S_REGISTER");
    return ret;
  }
  
  ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, reg);
  if (ret < 0) {
    perror("ioctl VIDIOC_DBG_G_REGISTER");
    return ret;
  }
  printf("%s (0x%x) NEW value = 0x%02x\n",
	 reg_name, (int) reg->reg, (int) reg->val);
  
  /* wait for sensor to adjust */
  usleep(500000);

  return 0;
}


/* disp_reg is for DEBUG ONLY */
void disp_reg(struct grab_context *gctx,
	    char *reg_name,
	    int reg_addr)
{
  int ret;
  struct v4l2_register reg;

  reg.reg = reg_addr;
  reg.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
  reg.match_chip = gctx->i2c_addr;
  ret = ioctl(gctx->vfd, VIDIOC_DBG_G_REGISTER, &reg);
  if (ret < 0)
    perror("ioctl VIDIOC_DBG_G_REGISTER");
  else
    printf("%s (0x%x) current value = 0x%02x\n",
	 reg_name, reg_addr, (int) reg.val);
}


#define BGGR(x,y,w) bggr[((x) & 1 ? 2*(x)+!((y)&1) : 2*(x)+((y)&1)) + \
			 ((y)>>1) * w * 4]

void convert_BGGR_to_RGB24_and_write(struct grab_context *gctx,
				     unsigned char *bggr,
				     FILE *f, int algo)
{
  int ii, jj, start;
  unsigned int val;
  unsigned char rgb[6];
  __u32 w;

  switch(algo) {
  case BGGR_2_RGB_ALGO_NNR:
    start = 0;
    break;
  case BGGR_2_RGB_ALGO_AVR:
    start = 1;
    break;
  default:
    printf("convert_BGGR_to_RGB24_and_write: unknown algorithm %d\n", algo);
    return;
  }

  /* borders are not included */
  w = gctx->width;
  for (ii=start; ii<gctx->height - 2; ii+=2) {

    for (jj=start; jj<gctx->width - 2; jj+=2) {
      if (algo == BGGR_2_RGB_ALGO_NNR) {
	/* line ii is BGBG... */

	/* RGB components of "center" pixel */
	rgb[0] = BGGR(jj+1, ii+1, w);
	val = BGGR(jj, ii+1, w) + BGGR(jj+1, ii, w);
	rgb[1] = val >> 1;
	rgb[2] = BGGR(jj, ii, w);

	/* RGB components of "right" pixel */
	rgb[3] = BGGR(jj+1, ii+1, w);
	val = BGGR(jj+2, ii+1, w) + BGGR(jj+1, ii, w);
	rgb[4] = val >> 1;
	rgb[5] = BGGR(jj+2, ii, w);
      }
      else if (algo == BGGR_2_RGB_ALGO_AVR) {
	/* line ii is GRGR... */

	/* RGB components of "red" pixel */
	rgb[0] = BGGR(jj, ii, w);
	val = BGGR(jj-1, ii, w) + BGGR(jj+1, ii, w) +
	  BGGR(jj, ii-1, w) + BGGR(jj, ii+1, w);
	rgb[1] = val >> 2;
	val = BGGR(jj-1, ii-1, w) + BGGR(jj+1, ii-1, w) +
	  BGGR(jj-1, ii+1, w) + BGGR(jj+1, ii+1, w);
	rgb[2] = val >> 2;

	/* RGB components of "green" pixel */
	val = BGGR(jj, ii, w) + BGGR(jj+2, ii, w);
	rgb[3] = val >> 1;
	rgb[4] = BGGR(jj+1, ii, w);
	val = BGGR(jj+1, ii-1, w) + BGGR(jj+1, ii+1, w);
	rgb[5] = val >> 1;
      }

      fwrite(rgb, sizeof(rgb), 1, f);
    }

    for (jj=start; jj<gctx->width - 2; jj+=2) {
      if (algo == BGGR_2_RGB_ALGO_NNR) {
	/* line ii+1 is GRGR... */

	/* RGB components of "center" pixel */
	rgb[0] = BGGR(jj+1, ii+1, w);
	val = BGGR(jj, ii+1, w) + BGGR(jj+1, ii+2, w);
	rgb[1] = val >> 1;
	rgb[2] = BGGR(jj, ii+2, w);

	/* RGB components of "right" pixel */
	rgb[3] = BGGR(jj+1, ii+1, w);
	val = BGGR(jj+1, ii+2, w) + BGGR(jj+2, ii+1, w);
	rgb[4] = val >> 1;
	rgb[5] = BGGR(jj+2, ii+2, w);
      }
      else if (algo == BGGR_2_RGB_ALGO_AVR) {
	/* line ii+1 is BGBG... */

	/* RGB components of "green" pixel */
	val = BGGR(jj, ii, w) + BGGR(jj, ii+2, w);
	rgb[0] = val >> 1;
	rgb[1] = BGGR(jj, ii+1, w);
	val = BGGR(jj-1, ii+1, w) + BGGR(jj+1, ii+1, w);
	rgb[2] = val >> 1;

	/* RGB components of "blue" pixel */
	val = BGGR(jj, ii, w) + BGGR(jj+2, ii, w) +
	  BGGR(jj, ii+2, w) + BGGR(jj+2, ii+2, w);
	rgb[3] = val >> 2;
	val = BGGR(jj, ii+1, w) + BGGR(jj+2, ii+1, w) +
	  BGGR(jj+1, ii, w) + BGGR(jj+1, ii+2, w);
	rgb[4] = val >> 2;
	rgb[5] = BGGR(jj+1, ii+1, w);
      }

      fwrite(rgb, sizeof(rgb), 1, f);
    }
  }
}

void convert_BGGR_to_Bayer_and_write(struct grab_context *gctx,
				     unsigned char *bggr, FILE *f)
{
  unsigned char *ptr, *ptr2, *end, *endl;
  unsigned char rgb[6];

  ptr = bggr;
  end = bggr + gctx->width * gctx->height * 2;

  while (ptr < end) {
    ptr2 = ptr;
    endl = ptr + gctx->width * 2;
    rgb[0] = 0; rgb[1] = 0; rgb[3] = 0; rgb[5] = 0;
    while (ptr2 < endl) {
      /* blue pixel */
      rgb[2] = *ptr2;
      /* green pixel */
      rgb[4] = *(ptr2 + 3);
      fwrite(rgb, sizeof(rgb), 1, f);
      ptr2 += 4;
    }

    ptr2 = ptr;
    rgb[0] = 0; rgb[2] = 0; rgb[4] = 0; rgb[5] = 0;
    while (ptr2 < endl) {
      /* green pixel */
      rgb[1] = *(ptr2 + 1);
      /* red pixel */
      rgb[3] = *(ptr2 + 2);
      fwrite(rgb, sizeof(rgb), 1, f);
      ptr2 += 4;
    }
    /* ignore next line because of data redundancy */
    ptr += gctx->width * 4;
  }
}

void convert_GBRP_to_Bayer_and_write(struct grab_context *gctx,
				     unsigned char *gp, unsigned char *bp,
				     unsigned char *rp, FILE *f)
{
  unsigned char *g, *b, *r, *endg;
  unsigned char rgb[6];
  int jj;

  g = gp; b = bp; r = rp;
  endg = gp + gctx->width * gctx->height;

  while (g < endg) {
    /* BGBG line */
    rgb[0] = 0; rgb[1] = 0; rgb[3] = 0; rgb[5] = 0;
    for (jj=0; jj<gctx->width; jj+=2) {
      /* blue pixel */
      rgb[2] = b[jj>>1];
      /* green pixel */
      rgb[4] = g[jj+1];
      fwrite(rgb, sizeof(rgb), 1, f);
    }

    /* GRGR line */
    rgb[0] = 0; rgb[2] = 0; rgb[4] = 0; rgb[5] = 0;
    for (jj=0; jj<gctx->width; jj+=2) {
      /* green pixel */
      rgb[1] = g[jj];
      /* red pixel */
      rgb[3] = r[jj>>1];
      fwrite(rgb, sizeof(rgb), 1, f);
    }

    /* ignore next line because of data redundancy */
    g += gctx->width * 2;
    b += gctx->width;
    r += gctx->width;
  }
}

int save_image(struct grab_context *gctx)
{
  int ret;
  int ii;
  char filename[256];
  __u32 height;

  ret = 0;

  switch (gctx->pixelformat) {

  case V4L2_PIX_FMT_UYVY:
  case V4L2_PIX_FMT_YUYV:
    {
      /* Packed YUV data format:
	 
	 OV7720 outputs data as follows (see datasheet table 1.2):
	 UYVY */

      FILE *fy, *fu, *fv;
      void *ptr;
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_Y_NAME);
      fy = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_U_NAME);
      fu = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_V_NAME);
      fv = fopen(filename, "wb");
      if ((fy == NULL) || (fu == NULL) || (fv == NULL)) {
	perror("fopen");
	ret = -1;
	goto save_UYVY_end;
      }
      printf("Saving images..."); fflush(stdout);
      height = gctx->height * gctx->nr_buffers;
      fprintf(fy, "P5 %d %d 255\n", gctx->width  , height);
      fprintf(fu, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(fv, "P5 %d %d 255\n", gctx->width/2, height);
      for (ii=0; ii<gctx->nr_buffers; ii++) {
	ptr = gctx->buffers[ii].start;
	while (ptr < gctx->buffers[ii].start + gctx->buffers[ii].length) {
	  if (gctx->pixelformat == V4L2_PIX_FMT_UYVY) {
	    fwrite(ptr+0, 1, 1, fu);
	    fwrite(ptr+1, 1, 1, fy);
	    fwrite(ptr+2, 1, 1, fv);
	    fwrite(ptr+3, 1, 1, fy);
	  }
	  else {
	    fwrite(ptr+0, 1, 1, fy);
	    fwrite(ptr+1, 1, 1, fu);
	    fwrite(ptr+2, 1, 1, fy);
	    fwrite(ptr+3, 1, 1, fv);
	  }
	  ptr += 4;
	}
      }
      printf("done.\n");
    
    save_UYVY_end:
      if (fy != NULL) fclose(fy);
      if (fu != NULL) fclose(fu);
      if (fv != NULL) fclose(fv);
    }
    break;

  case V4L2_PIX_FMT_YUV422P:
    {
      FILE *fy, *fu, *fv;
      void *ptr;

      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_Y_NAME);
      fy = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_U_NAME);
      fu = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_V_NAME);
      fv = fopen(filename, "wb");
      if ((fy == NULL) || (fu == NULL) || (fv == NULL)) {
	perror("fopen");
	ret = -1;
	goto save_YUV422P_end;
      }
      printf("Saving images..."); fflush(stdout);
      height = gctx->height * gctx->nr_buffers;
      fprintf(fy, "P5 %d %d 255\n", gctx->width, height);
      fprintf(fu, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(fv, "P5 %d %d 255\n", gctx->width/2, height);
      for (ii=0; ii<gctx->nr_buffers; ii++) {
	ptr = gctx->buffers[ii].start;
	printf("Y plane start = 0x%08x\n", ptr);
	fwrite(ptr, 1, gctx->width*gctx->height, fy);
	ptr += page_align(gctx->width * gctx->height);
	printf("Cb/U plane start = 0x%08x\n", ptr);
	fwrite(ptr, 1, gctx->width*gctx->height/2, fu);
	ptr += page_align(gctx->width * gctx->height / 2);
	printf("Cr/V plane start = 0x%08x\n", ptr);
	fwrite(ptr, 1, gctx->width*gctx->height/2, fv);
      }
      printf("done.\n");

    save_YUV422P_end:
      if (fy != NULL) fclose(fy);
      if (fu != NULL) fclose(fu);
      if (fv != NULL) fclose(fv);
    }
    break;

  case V4L2_PIX_FMT_SBGGR8:
    {
      /* Packed raw Bayer data format:
	 
	 OV7720 outputs data as follows (see datasheet 0.3.4):
	 odd lines (sensor odd lines are BGBG...):
	 B11 G21 R22 G12 B13 G23 R24 G14 ...
	 even lines (sensor even lines are GRGR...):
	 B31 G21 R22 G32 B33 G33 R24 G34 ...

	 An odd line and its successor have same R data.
	 An even line and its successor have same B data.
	 Each line has different G data but all G data can be retrieved
	 from 2 successive odd or even lines.
	 Because of this data redundancy, all image data can be read and
	 converted using only odd or even lines.

	 Below odd lines are used (odd as defined in OV7720 datasheet where
	 lines and columns count starts at 1, means even here with
	 count starting at 0 ...).
      */

      FILE *fg, *fr, *fb, *fi, *f;
      void *ptr, *ptr2, *end, *endl;
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_G_NAME);
      fg = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_R_NAME);
      fr = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_B_NAME);
      fb = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_NAME);
      fi  = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_BAYER_NAME);
      f  = fopen(filename, "wb");
      if ((fg == NULL) || (fr == NULL) || (fb == NULL) || (fi == NULL) ||
	  (f == NULL)) {
	perror("fopen");
	ret = -1;
	goto save_SBGGR8_end;
      }
      printf("Saving images..."); fflush(stdout);
      height = gctx->height * gctx->nr_buffers;
      fprintf(fg, "P5 %d %d 255\n", gctx->width  , height);
      fprintf(fr, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(fb, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(f , "P6 %d %d 255\n", gctx->width  , height);
      fprintf(fi, "P6 %d %d 255\n", gctx->width-2,
	      (gctx->height-2) * gctx->nr_buffers);
      for (ii=0; ii<gctx->nr_buffers; ii++) {
	ptr = gctx->buffers[ii].start;
	end = ptr + gctx->buffers[ii].length;
	while (ptr < end) {
	  fwrite(ptr+0, 1, 1, fb);
	  fwrite(ptr+1, 1, 1, fg);
	  fwrite(ptr+2, 1, 1, fr);
	  fwrite(ptr+3, 1, 1, fg);
	  ptr += 4;
	}

	convert_BGGR_to_RGB24_and_write(gctx, gctx->buffers[ii].start, fi,
					BGGR_2_RGB_ALGO_NNR);

	/* pseudo Bayer image */
	convert_BGGR_to_Bayer_and_write(gctx, gctx->buffers[ii].start, f);
      }
      printf("done.\n");

    save_SBGGR8_end:
      if (fg != NULL) fclose(fg);
      if (fr != NULL) fclose(fr);
      if (fb != NULL) fclose(fb);
      if (fi != NULL) fclose(fi);
      if (f  != NULL) fclose(f );
    }
    break;

  case V4L2_PIX_FMT_GRB422P:
    {
      /* Planarized raw Bayer data format:
	 
	 The planarized image buffer contains R, G and B components in
	 different memory areas called planes. For memory management reasons,
	 each plane size is an integer multiple of memory page.

	 G plane size is width * height. It starts at the base address
	 of the image buffer which is always aligned with a page boundary.

	 Next is B plane which size is (width/2) * height. It starts at the
	 page boundary following the end of the G plane.

	 Next is R plane which size is (width/2) * height. It starts at the
	 page boundary following the end of the B plane.


	 OV7720 outputs data as follows (see datasheet 0.3.4):
	 odd lines (sensor odd lines are BGBG...):
	 B11 G21 R22 G12 B13 G23 R24 G14 ...
	 even lines (sensor even lines are GRGR...):
	 B31 G21 R22 G32 B33 G33 R24 G34 ...

	 The planarized acquisition stores pixel values as follows:
	   . G plane:
	   G21 G12 G23 G14 G25 G16 ...
	   G21 G32 G23 G34 G25 G36 ...
	   G41 G32 G43 G34 G45 G36 ...
	   G41 G52 G43 G54 G45 G56 ...
	   ...
	   Each line contains width G values. Each line has different G data
	   but all G data can be retrieved from 2 successive odd or even lines.

	   . B plane:
	   B11 B13 B15 ...
	   B31 B33 B35 ...
	   B31 B33 B35 ...
	   B51 B53 B55 ...
	   B51 B53 B55 ...
	   ...
	   Each line contains width/2 B values. Each line is duplicated 
	   because blue pixels are present only in odd sensor lines.
	   => Consider only odd or even lines.

	   . R plane:
	   R22 R24 R26 ...
	   R22 R24 R26 ...
	   R42 R44 R46 ...
	   R42 R44 R46 ...
	   ...
	   Each line contains width/2 R values. Each line is duplicated 
	   because red pixels are present only in even sensor lines.
	   => Consider only odd or even lines.

      */
      FILE *fg, *fr, *fb, *f;
      void *ptr, *ptrB, *ptrR;
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_G_NAME);
      fg = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_R_NAME);
      fr = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_B_NAME);
      fb = fopen(filename, "wb");
      sprintf(filename, "%s/%s", IMAGE_DIR, IMAGE_BAYER_NAME);
      f  = fopen(filename, "wb");
      if ((fg == NULL) || (fr == NULL) || (fb == NULL) || (f == NULL)) {
	perror("fopen");
	ret = -1;
	goto save_GRB422_end;
      }
      printf("Saving images..."); fflush(stdout);
      height = gctx->height * gctx->nr_buffers;
      fprintf(fg, "P5 %d %d 255\n", gctx->width  , height);
      fprintf(fr, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(fb, "P5 %d %d 255\n", gctx->width/2, height);
      fprintf(f , "P6 %d %d 255\n", gctx->width  , height);
      for (ii=0; ii<gctx->nr_buffers; ii++) {
	ptr = gctx->buffers[ii].start;
	ptrB = ptr  + page_align(gctx->width * gctx->height);
	ptrR = ptrB + page_align(gctx->width * gctx->height / 2);
	printf("G plane start = 0x%08x\n", ptr);
	fwrite(ptr, 1, gctx->width*gctx->height, fg);
	printf("B plane start = 0x%08x\n", ptrB);
	fwrite(ptrB, 1, gctx->width*gctx->height/2, fb);
	printf("R plane start = 0x%08x\n", ptrR);
	fwrite(ptrR, 1, gctx->width*gctx->height/2, fr);

	/* pseudo Bayer image */
	convert_GBRP_to_Bayer_and_write(gctx, ptr, ptrB, ptrR, f);
      }
      printf("done.\n");

    save_GRB422_end:
      if (fg != NULL) fclose(fg);
      if (fr != NULL) fclose(fr);
      if (fb != NULL) fclose(fb);
      if (f  != NULL) fclose(f );
    }
    break;

  default:
    printf("??? %d unknown pixel format.\n", gctx->pixelformat);
    ret = -1;
  }
 	  
  return ret;
}

int grab_start(struct grab_context *gctx)
{
  int ret;
  int ii;
  struct v4l2_buffer buffer;

  memset(&buffer, 0, sizeof (buffer));
  buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buffer.memory = V4L2_MEMORY_MMAP;

  for (ii=0; ii<gctx->nr_buffers; ii++) {
    buffer.index = ii;
    ret = ioctl(gctx->vfd, VIDIOC_QBUF, &buffer);
    if (ret < 0) {
      perror ("ioctl VIDIOC_QBUF");
      return ret;
    }

    printf("buffer %u: bytesused = %u\n", buffer.index, buffer.bytesused);
  }

  ret = ioctl(gctx->vfd, VIDIOC_STREAMON, &buffer.type);
  if (ret < 0) {
    perror ("ioctl VIDIOC_STREAMON");
    return ret;
  }
  
  printf("Wait for sensor auto adjust..."); 
  fflush(stdout);
  sleep(1);
  printf(" Done.\n");

  return 0;
}

int grab_stop(struct grab_context *gctx)
{
  int ret;
  enum v4l2_buf_type type;

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

  ret = ioctl(gctx->vfd, VIDIOC_STREAMOFF, &type);
  if (ret < 0) {
    perror ("ioctl VIDIOC_STREAMOFF");
    return ret;
  }

  return 0;
}

int grab_image(struct grab_context *gctx)
{
  int ret;
  int ii;
  struct v4l2_buffer buffer;
  
  /* start capture */
  ret = grab_start(gctx);
  if (ret < 0)
    return ret;

  buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buffer.memory = V4L2_MEMORY_MMAP;

  for (ii=0; ii<gctx->nr_buffers; ii++) {
    /* retrieve image when sensor turned on */
    buffer.index = ii;
    ret = ioctl(gctx->vfd, VIDIOC_DQBUF, &buffer);
    if (ret < 0) {
      if (errno == EAGAIN)
	printf("Image not grabbed yet.\n");
      else
	perror("ioctl VIDIOC_DQBUF");
      return ret;
    }
  }

  /* queue all buffers first */
  for (ii=0; ii<gctx->nr_buffers; ii++) {
    buffer.index = ii;
    ret = ioctl(gctx->vfd, VIDIOC_QBUF, &buffer);
    if (ret < 0) {
      perror ("ioctl VIDIOC_QBUF");
      goto stream_off;
    }
    printf("buffer %u: bytesused = %u\n", buffer.index, buffer.bytesused);
  }

  /* dequeue buffers */
  for (ii=0; ii<gctx->nr_buffers; ii++) {
    buffer.index = ii;
    ret = ioctl(gctx->vfd, VIDIOC_DQBUF, &buffer);
    if (ret < 0) {
      if (errno == EAGAIN)
	printf("Image not grabbed yet.\n");
      else
	perror("ioctl VIDIOC_DQBUF");
      goto stream_off;
    }
    printf("%ld.%06ld DQBUF %02d end, frame %u.\n",
	   buffer.timestamp.tv_sec, buffer.timestamp.tv_usec,
	   buffer.index, buffer.sequence);
/*     disp_reg(gctx, "AECH", OV7720_AECH); */
/*     disp_reg(gctx, "AEC" , OV7720_AEC ); */
/*     disp_reg(gctx, "AGC" , OV7720_AGC ); */
/*     disp_reg(gctx, "COM8", OV7720_COM8); */
/*     disp_reg(gctx, "COM9", OV7720_COM9); */
  }

  ret = save_image(gctx);
  if (ret < 0)
    goto stream_off;

 stream_off:
  /* stop capture */
  ret = grab_stop(gctx);
  if (ret < 0)
    return ret;

  return 0;
}


int grab_init(struct grab_context *gctx)
{
  int ret;
  int input;
  int ii;
  struct v4l2_register reg;
  __u64 mask;
  struct v4l2_control ctrl;
  struct v4l2_format format;
  struct v4l2_requestbuffers reqbuf;
  struct v4l2_buffer buffer;

  gctx->vfd = -1;
  gctx->buffers = NULL;
  gctx->i2c_addr = OV7720_I2C_ADDR;

  gctx->vfd = open(VIDEO_DEV_NAME, O_RDWR);
  if (gctx->vfd < 0) {
    printf("%s ", VIDEO_DEV_NAME);
    perror("open failed");
    return -1;
  }

  printf("%s open succeeded.\n", VIDEO_DEV_NAME);

#if 0
  /* Clock prescaler: CLKRC
     CLKRC<6> = 1: no prescale, PCLK = MCLK
     CLKRC<6> = 0: prescale defined by PCLK = 2*MCLK/(CLKRC<5:0>+1)
     examples:
     CLKRC<6:0> = 0x00 => fastest rate: PCLK = 2*MCLK
     CLKRC<6:0> = 0x01 => PCLK = MCLK
     CLKRC<6:0> = 0x02 => PCLK = MCLK/2
     CLKRC<6:0> = 0x03 => PCLK = MCLK*2/3
     CLKRC<6:0> = 0x3f => slowest rate: PCLK = MCLK/16 */
  reg.reg = OV7720_CLKRC;
  mask = 0x8f; /* bits <6:0> */
  reg.val = 0x04;
  set_reg(gctx, "CLKRC", &reg, &mask);
#endif
  
#if 0
  /* Output drive capability: COM2<1:0>
     00: Output drive x 1
     01: Output drive x 2
     10: Output drive x 3
     11: Output drive x 4 */
  reg.reg = OV7720_COM2;
  reg.val = 0x00;  /* between 0x00 and 0x03 */
  mask    = 0x03;  /* bits <1:0> */
  set_reg(gctx, "COM2", &reg, &mask);
#endif

#if 0
  /* Output control:
     COM3<5> B/R swap in RGB output mode
     COM3<4> Y/UV swap in YUV output mode, G/RB swap in RGB output mode
     COM3<0> Sensor color bar test pattern output enable */
  reg.reg = OV7720_COM3;
  reg.val = 0;
  reg.val |= 0x10;  /* B/R swap */
  reg.val |= 0x08;  /* Y/UV swap */
  mask = 0x19;
  set_reg(gctx, "COM3", &reg, &mask);
#endif

#if 0
  reg.reg = OV7720_HSTART;
  reg.val = 0x26;
  mask = 0xff;
  set_reg(gctx, "HSTART", &reg, &mask);
  reg.reg = OV7720_HREF;
  reg.val = 0x08;
  mask = 0x18;
  set_reg(gctx, "HREF", &reg, &mask);
#endif
#if 0
  reg.reg = OV7720_VSTART;
  reg.val = 0x07;
  mask = 0xff;
  set_reg(gctx, "VSTART", &reg, &mask);
#endif
  disp_reg(gctx, "AECH", OV7720_AECH);
  disp_reg(gctx, "AEC" , OV7720_AEC );
  disp_reg(gctx, "AGC" , OV7720_AGC );
  disp_reg(gctx, "COM8", OV7720_COM8);
  disp_reg(gctx, "COM9", OV7720_COM9);
#if 0
  reg.reg = OV7720_LUMHTH;
  reg.val = 0x20;
  mask = 0xff;
  set_reg(gctx, "LumHTh", &reg, &mask);
  reg.reg = OV7720_LUMLTH;
  reg.val = 0x00;
  mask = 0xff;
  set_reg(gctx, "LumLTh", &reg, &mask);
#endif

  /* select input 0, the only one */
  input = 0;

  ret = ioctl(gctx->vfd, VIDIOC_S_INPUT, &input);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_INPUT");
    return ret;
  }

  /* select vga/qvga format */
  ctrl.id = OV7720_CID_QVGA;
  if (gctx->width == 640)
    ctrl.value = 0;
  else
    ctrl.value = 1;

  ret = ioctl(gctx->vfd, VIDIOC_S_CTRL, &ctrl);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_CTRL OV7720_CID_QVGA");
    return ret;
  }

  /* configure acquisition format */
  format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  format.fmt.pix.width = gctx->width;
  format.fmt.pix.height = gctx->height;
  format.fmt.pix.pixelformat = gctx->pixelformat;
  format.fmt.pix.field = V4L2_FIELD_NONE;

  ret = ioctl(gctx->vfd, VIDIOC_S_FMT, &format);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_FMT");
    return ret;
  }

  /* request buffers to store grabbed image */
  reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqbuf.memory = V4L2_MEMORY_MMAP;
  reqbuf.count = gctx->nr_buffers;

  ret = ioctl(gctx->vfd, VIDIOC_REQBUFS, &reqbuf);
  if (ret < 0) {
    if (errno == EINVAL)
      printf ("Video capturing or mmap-streaming is not supported\n");
    else
      perror ("ioctl VIDIOC_REQBUFS");
    return ret;
  }

  if (reqbuf.count < gctx->nr_buffers) {
    printf("   !!! Only %d buffers available\n", reqbuf.count);
    gctx->nr_buffers = reqbuf.count;
  }

  gctx->buffers = calloc(gctx->nr_buffers, sizeof (*gctx->buffers));
  if (gctx->buffers == NULL) {
    perror("calloc gctx->buffers");
    return -1;
  }

  for (ii=0; ii<gctx->nr_buffers; ii++) {
    /* query buffer to allocate and mmap it */
    memset(&buffer, 0, sizeof (buffer));
    buffer.type = reqbuf.type;
    buffer.memory = V4L2_MEMORY_MMAP;
    buffer.index = ii;

    ret = ioctl(gctx->vfd, VIDIOC_QUERYBUF, &buffer);
    if (ret < 0) {
      perror ("ioctl VIDIOC_QUERYBUF");
      exit (EXIT_FAILURE);
    }

    printf("buffer %u: bytesused = %u\n",
	   buffer.index, buffer.bytesused);

    gctx->buffers[ii].length = buffer.length; /* remember for munmap() */
    gctx->buffers[ii].start = mmap (NULL, buffer.length,
				    PROT_READ | PROT_WRITE,
				    MAP_SHARED,
				    gctx->vfd, buffer.m.offset);
    if (MAP_FAILED == gctx->buffers[ii].start) {
      perror ("mmap");
      return -1;
    }
  }

  return 0;
}

void grab_end(struct grab_context *gctx)
{
  int ii;
  if (gctx->buffers != NULL) {
    for (ii=0; ii<gctx->nr_buffers; ii++) {
      if ((gctx->buffers[ii].start != MAP_FAILED) &&
	  (gctx->buffers[ii].length > 0))
	munmap (gctx->buffers[ii].start, gctx->buffers[ii].length);
    }
    
    free(gctx->buffers);
  }

  if (gctx->vfd != -1) close(gctx->vfd);
}


#define idx_BGGR(x,y,w) ((x) & 1 ? 2*(x)+!((y)&1) : 2*(x)+((y)&1)) + \
			 ((y)>>1) * w * 4

int main(int argc, char *argv[])
{
  int video_dev;
  int ret;
  struct grab_context gctx;
  struct v4l2_input input;
  
  printf("\n\n   ***   test_grab   ***\n\n");
  printf("Basic image grab with V4L2 compatible PXA camera driver\n\n");
  usleep(100000); /* for proper printing */

  ret = check_args(argc, argv, &gctx);
  if (ret) exit(1);

  ret = grab_init(&gctx);
  if (ret < 0) goto end;

  ret = grab_image(&gctx);
  if (ret < 0) goto end;
/*   frame_drop(&gctx); */

 end:
  grab_end(&gctx);

  exit(0);
}
