#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/videodev2.h>
#include <linux/ov7720.h>

#define VIDEO_DEV_NAME "/dev/video0"

int query_capabilities(int fd)
{
  int ret;
  struct v4l2_capability cap;

  ret = ioctl(fd, VIDIOC_QUERYCAP, &cap);
  if (ret != 0) {
    perror("ioctl VIDIOC_QUERYCAP");
    return ret;
  }

  printf("\n\n");
  printf("DEVICE CAPABILITIES:\n");
  printf("--------------------\n");
  printf("driver      : %s\n", cap.driver);
  printf("card        : %s\n", cap.card);
  printf("bus_info    : %s\n", cap.bus_info);
  printf("version     : %u.%u.%u\n",
        (cap.version >> 16) & 0xFF,
        (cap.version >> 8) & 0xFF,
         cap.version & 0xFF);
  printf("capabilities: 0x%08x\n", cap.capabilities);
  if (cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)
    printf("  Is a video capture device\n");
  if (cap.capabilities & V4L2_CAP_STREAMING)
    printf("  streaming I/O ioctls\n");
  usleep(100000); /* for proper printing */

  return 0;
}

int query_input(int fd, struct v4l2_input *input)
{
  int ret;
  int index;

  ret = ioctl (fd, VIDIOC_G_INPUT, &index);
  if (ret != 0) {
    perror("ioctl VIDIOC_G_INPUT");
    return ret;
  }

  memset (input, 0, sizeof (*input));
  input->index = index;
  ret = ioctl (fd, VIDIOC_ENUMINPUT, input);
  if (ret != 0) {
    perror("ioctl VIDIOC_ENUMINPUT");
    return ret;
  }

  printf("\n\n");
  printf("CURRENT INPUT %d:\n", index);
  printf("----------------\n");
  printf("name        : %s\n", input->name);
  printf("type        : ");
  if (input->type == V4L2_INPUT_TYPE_TUNER) printf("tuner\n");
  else if (input->type == V4L2_INPUT_TYPE_CAMERA) printf("camera\n");
  else printf("unknown!\n");
  printf("standard    : 0x%08x\n", input->std);
  printf("status      : 0x%08x\n", input->status);
  usleep(100000); /* for proper printing */

  return 0;
}

int query_standards(int fd, struct v4l2_input *input)
{
  int ret;
  struct v4l2_standard standard;

  memset (&standard, 0, sizeof (standard));
  standard.index = 0;
  while (1) {
    ret = ioctl(fd, VIDIOC_ENUMSTD, &standard);
    if (ret == 0) {
      if (standard.id & input->std)
	printf ("%s\n", standard.name);
    }
    else {
      /* EINVAL indicates the end of the enumeration, which cannot be
	 empty unless this device falls under the USB exception. */
      if (errno != EINVAL) perror("ioctl VIDIOC_ENUMSTD");
      break;
    }
    standard.index++;
  }
  usleep(100000); /* for proper printing */

  return 0;
}

int query_formats(int fd)
{
  int ret;
  struct v4l2_fmtdesc fmtdesc;
  struct v4l2_format format;
  char fourcc_str[sizeof(fmtdesc.pixelformat)+1];

  printf("\n\n");
  printf("POSSIBLE IMAGE FORMATS:\n");
  printf("-----------------------\n");
  memset (&fmtdesc, 0, sizeof (fmtdesc));
  fmtdesc.index = 0;
  while (1) {
    fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    ret = ioctl(fd, VIDIOC_ENUM_FMT, &fmtdesc);
    if (ret == 0) {
      printf ("%s:\n", fmtdesc.description);
      printf("  flags      : 0x%08x\n", fmtdesc.flags);
      strncpy(fourcc_str,
	      (char*) &fmtdesc.pixelformat,
	      sizeof(fmtdesc.pixelformat));
      fourcc_str[sizeof(fmtdesc.pixelformat)] = 0;
      printf("  pixelformat: %s 0x%08x\n", fourcc_str, fmtdesc.pixelformat);
      usleep(100000); /* for proper printing */
      
      /* try this format */
      format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      format.fmt.pix.width = 640;
      format.fmt.pix.height = 480;
      format.fmt.pix.pixelformat = fmtdesc.pixelformat;
      format.fmt.pix.field = V4L2_FIELD_NONE;
      format.fmt.pix.bytesperline = 0;
      ret = ioctl(fd, VIDIOC_TRY_FMT, &format);
      printf("\n  VIDIOC_TRY_FMT (VGA size, progressive scan) ");
      if (ret == 0) printf("successful ");
      else printf("failed (%d %d) ", ret, errno);
      printf("for this format.\n");
      usleep(100000); /* for proper printing */

      /* set this format */
      if (ret == 0) {
	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	ret = ioctl(fd, VIDIOC_G_FMT, &format);
	if (ret != 0) perror("ioctl VIDIOC_G_FMT");
	else {
	  printf("\n  VIDIOC_G_FMT Current format:\n");
	  printf("    width        : %u\n", format.fmt.pix.width);
	  printf("    height       : %u\n", format.fmt.pix.height);
	  strncpy(fourcc_str,
		  (char*) &format.fmt.pix.pixelformat,
		  sizeof(format.fmt.pix.pixelformat));
	  fourcc_str[sizeof(format.fmt.pix.pixelformat)] = 0;
	  printf("    pixelformat  : %s 0x%08x\n",
		 fourcc_str, format.fmt.pix.pixelformat);
	  printf("    field        : %d\n", format.fmt.pix.field);
	  printf("    bytesperline : %u\n", format.fmt.pix.bytesperline);
	  printf("    sizeimage    : %u\n", format.fmt.pix.sizeimage);
	  printf("    colorspace   : %d\n", format.fmt.pix.colorspace);
	  usleep(100000); /* for proper printing */

	  format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	  format.fmt.pix.width = 640;
	  format.fmt.pix.height = 480;
	  format.fmt.pix.pixelformat = fmtdesc.pixelformat;
	  format.fmt.pix.field = V4L2_FIELD_NONE;
	  format.fmt.pix.bytesperline = 0;
	  ret = ioctl(fd, VIDIOC_S_FMT, &format);
	  strncpy(fourcc_str,
		  (char*) &format.fmt.pix.pixelformat,
		  sizeof(format.fmt.pix.pixelformat));
	  fourcc_str[sizeof(format.fmt.pix.pixelformat)] = 0;
	  printf("\n  VIDIOC_S_FMT Set %s format: ", fourcc_str);
	  fflush(stdout);
	  if (ret != 0) perror("ioctl VIDIOC_S_FMT");
	  else printf("successful\n");
	  usleep(100000); /* for proper printing */
	}
      }
    }
    else {
      /* EINVAL indicates the end of the enumeration. */
      if (errno != EINVAL) perror("ioctl VIDIOC_ENUM_FMT");
      break;
    }
    printf("\n");
    fmtdesc.index++;
  }

  return 0;
}

void print_queryctrl(struct v4l2_queryctrl *qc)
{
  printf("\n%s:\n", qc->name);
  printf("  type % 2d, min % 3d, max % 3d, step % 3d, default % 3d (0x%02x), flags 0x%08x\n",
	 qc->type, qc->minimum, qc->maximum, qc->step,
	 qc->default_value, qc->default_value, qc->flags);
  usleep(100000); /* for proper printing */
}

int get_ctrl_val(int fd, __u32 id)
{
  int ret;
  struct v4l2_control ctrl;

  ctrl.id = id;
  ret = ioctl(fd, VIDIOC_G_CTRL, &ctrl);
  if (ret == 0)
    printf("  current value: 0x%02x\n", ctrl.value);
  else {
    if (errno == EINVAL) {
      printf("  No value for this control.\n");
      ret = 0;
    }
    else
      perror("ioctl VIDIOC_G_CTRL");
  }
  usleep(100000); /* for proper printing */

  return ret;
}

int query_controls(int fd)
{
  int ret, ii;
  struct v4l2_queryctrl queryctrl;

  printf("\n\n");
  printf("AVAILABLE CONTROLS:\n");
  printf("-------------------\n");

  for (queryctrl.id = V4L2_CID_BASE;
       queryctrl.id < V4L2_CID_LASTP1;
       queryctrl.id++) {
    ret = ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl);
    if (ret == 0) {
      if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
	continue;

      print_queryctrl(&queryctrl);

      /* get current value */
      ret = get_ctrl_val(fd, queryctrl.id);
      if (ret) return ret;
    }
    else {
      if (errno == EINVAL)
	continue;
      perror ("ioctl VIDIOC_QUERYCTRL");
      return ret;
    }
  }

  for (queryctrl.id = V4L2_CID_PRIVATE_BASE;;
       queryctrl.id++) {
    ret = ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl);
    if (ret == 0) {
      if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
	continue;

      print_queryctrl(&queryctrl);

      /* get current value */
      ret = get_ctrl_val(fd, queryctrl.id);
      if (ret) return ret;
    }
    else {
      if (errno == EINVAL)
	break;

      perror ("ioctl VIDIOC_QUERYCTRL");
      return ret;
    }
  }

  return 0;
}

int test_g_s_crop(int fd)
{
  int ret;
  struct v4l2_crop crop, new_crop;

  crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ret = ioctl(fd, VIDIOC_G_CROP, &crop);
  if (ret != 0) {
    perror("ioctl VIDIOC_G_CROP");
    return ret;
  }
  printf("  current window: top left corner (%d,%d), size (%d,%d)\n",
	 crop.c.left, crop.c.top,
	 crop.c.width, crop.c.height);
  printf("\n  setting crop region: size = 1/4 current size, centered:\n");
  new_crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  new_crop.c.width = crop.c.width/2;
  new_crop.c.height = crop.c.height/2;
  new_crop.c.left = crop.c.left + crop.c.width/4;
  new_crop.c.top = crop.c.top + crop.c.height/4;
  printf("  top left corner (%d,%d), size (%d,%d)\n",
	 new_crop.c.left, new_crop.c.top,
	 new_crop.c.width, new_crop.c.height);
  ret = ioctl(fd, VIDIOC_S_CROP, &new_crop);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_CROP");
    return ret;
  }
  crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ret = ioctl(fd, VIDIOC_G_CROP, &crop);
  if (ret != 0) {
    perror("ioctl VIDIOC_G_CROP");
    return ret;
  }
  printf("  current window: top left corner (%d,%d), size (%d,%d)\n",
	 crop.c.left, crop.c.top,
	 crop.c.width, crop.c.height);

  usleep(100000); /* for proper printing */

  return 0;
}

int print_cropcap(int fd)
{
  int ret;
  struct v4l2_cropcap cropcap;

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ret = ioctl(fd, VIDIOC_CROPCAP, &cropcap);
  if (ret < 0) {
    perror("ioctl VIDIOC_CROPCAP");
    return ret;
  }
  printf("  max rect (bounds): top left corner (%d,%d), size (%d,%d)\n",
	 cropcap.bounds.left, cropcap.bounds.top,
	 cropcap.bounds.width, cropcap.bounds.height);
  printf("  default rect: top left corner (%d,%d), size (%d,%d)\n",
	 cropcap.defrect.left, cropcap.defrect.top,
	 cropcap.defrect.width, cropcap.defrect.height);
  usleep(100000); /* for proper printing */

  return 0;
}

int query_crop(int fd)
{
  int ret, ii;
  struct v4l2_control ctrl;

  printf("\n\n");
  printf("CROP CAPABILITIES:\n");
  printf("------------------\n");

  ctrl.id = OV7720_CID_QVGA;
  ctrl.value = 0;
  ret = ioctl(fd, VIDIOC_S_CTRL, &ctrl);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_CTRL OV7720_CID_QVGA");
    return ret;
  }
  printf("\nVGA output format:\n");
  ret = print_cropcap(fd);
  if (ret < 0)
    return ret;

  ret = test_g_s_crop(fd);
  if (ret < 0)
    return ret;

  ctrl.id = OV7720_CID_QVGA;
  ctrl.value = 1;
  ret = ioctl(fd, VIDIOC_S_CTRL, &ctrl);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_CTRL OV7720_CID_QVGA");
    return ret;
  }
  printf("\nQVGA output format:\n");
  ret = print_cropcap(fd);
  if (ret < 0)
    return ret;

  ret = test_g_s_crop(fd);
  if (ret < 0)
    return ret;

  return 0;
}

int get_chip_ident(int fd)
{
  int ret;
  struct v4l2_chip_ident ident;

  ident.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
  ident.match_chip = 0x21;
  ret = ioctl(fd, VIDIOC_G_CHIP_IDENT, &ident);
  if (ret < 0) {
    perror("ioctl VIDIOC_G_CHIP_IDENT");
    return ret;
  }
  printf("\n\nCHIP ID:\n");
  printf("--------\n");
  printf("  ident   : %x\n", ident.ident);
  printf("  revision: %x\n", ident.revision);
  usleep(100000); /* for proper printing */

  return 0;
}

int reg_access(int fd)
{
  int ret;
  int addr, val;
  struct v4l2_register reg;
  
  printf("\n\nREGISTER ACCESS:\n");
  printf("----------------\n");
  while (1) {
    printf("\n  WARNING: DO NOT EDIT SENSOR REGISTERS UNLESS YOU KNOW \
WHAT YOU ARE DOING!\n");
    printf("  Enter address of register to modify, -1 to quit: ");
    fflush(stdout);
    scanf("%x", &addr);
    if (addr < 0)
      break;
    else if (addr > 0xff) {
      printf("  Invalid address\n");
      continue;
    }
    reg.match_type = V4L2_CHIP_MATCH_I2C_ADDR;
    reg.match_chip = 0x21;
    reg.reg = addr;
    ret = ioctl(fd, VIDIOC_DBG_G_REGISTER, &reg);
    if (ret < 0) {
      perror("ioctl VIDIOC_DBG_G_REGISTER");
      return ret;
    }
    val = reg.val;
    printf("  Current value: 0x%02x\n", val);
    printf("  Enter new value for register 0x%02x, -1 to quit: ", addr);
    fflush(stdout);
    scanf("%x", &val);
    if (val < 0)
      break;
    else if (val > 0xff) {
      printf("  Invalid value\n");
      continue;
    }
    reg.val = val;
    ret = ioctl(fd, VIDIOC_DBG_S_REGISTER, &reg);
    if (ret < 0) {
      perror("ioctl VIDIOC_DBG_S_REGISTER");
      return ret;
    }
    printf("  Done.\n");
  }

  return 0;
}

int prepare_test_buffer_mapping(int fd, int qvga, __u32 pixelformat)
{
  int ret;
  struct v4l2_control ctrl;
  struct v4l2_format format;

  ctrl.id = OV7720_CID_QVGA;
  ctrl.value = qvga;
  ret = ioctl(fd, VIDIOC_S_CTRL, &ctrl);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_CTRL OV7720_CID_QVGA");
    return ret;
  }
  format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  format.fmt.pix.width = 640;
  format.fmt.pix.height = 480;
  format.fmt.pix.pixelformat = pixelformat;
  format.fmt.pix.field = V4L2_FIELD_NONE;
  ret = ioctl(fd, VIDIOC_S_FMT, &format);
  if (ret != 0) {
    perror("ioctl VIDIOC_S_FMT");
    return ret;
  }
	  
  return 0;
}

int test_buffer_mapping(int fd)
{
  int ret;
  struct v4l2_requestbuffers reqbuf;
  struct {
    void *start;
    size_t length;
  } *buffers;
  unsigned int i;
  struct v4l2_buffer buffer;
  
  printf("\n\nTEST BUFFER MAPPING:\n");
  printf("--------------------\n");
  memset (&reqbuf, 0, sizeof (reqbuf));
  reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqbuf.memory = V4L2_MEMORY_MMAP;
  reqbuf.count = 2;

  ret = ioctl(fd, VIDIOC_REQBUFS, &reqbuf);
  if (ret < 0) {
    if (errno == EINVAL)
      printf ("Video capturing or mmap-streaming is not supported\n");
    else
      perror ("VIDIOC_REQBUFS");
    return ret;
  }

  printf("  %d buffers available.\n", reqbuf.count);

  buffers = calloc (reqbuf.count, sizeof (*buffers));
  for (i = 0; i < reqbuf.count; i++) {
    memset(&buffer, 0, sizeof (buffer));
    buffer.type = reqbuf.type;
    buffer.memory = V4L2_MEMORY_MMAP;
    buffer.index = i;

    ret = ioctl(fd, VIDIOC_QUERYBUF, &buffer);
    if (ret < 0) {
      perror ("VIDIOC_QUERYBUF");
      break;
    }

    printf("\n  Buffer %d:\n", buffer.index);
    printf("    type     : %d\n", buffer.type);
    printf("    bytesused: %u\n", buffer.bytesused);
    printf("    flags    : %u\n", buffer.flags);
    printf("    field    : %u\n", buffer.field);
    printf("    timestamp: %ld.%06ld\n",
	   buffer.timestamp.tv_sec, buffer.timestamp.tv_usec);
    printf("    sequence : %u\n", buffer.field);
    usleep(100000); /* for proper printing */

    buffers[i].length = buffer.length; /* remember for munmap() */

    buffers[i].start = mmap (NULL, buffer.length,
			     PROT_READ | PROT_WRITE, /* recommended */
			     MAP_SHARED,             /* recommended */
			     fd, buffer.m.offset);

    if (MAP_FAILED == buffers[i].start) {
      perror ("mmap");
      break;
    }

    ret = ioctl(fd, VIDIOC_QUERYBUF, &buffer);
    if (ret < 0) {
      perror ("VIDIOC_QUERYBUF");
      break;
    }

    printf("\n  Buffer %d after mmap call:\n", buffer.index);
    printf("    type     : %d\n", buffer.type);
    printf("    bytesused: %u\n", buffer.bytesused);
    printf("    flags    : %u\n", buffer.flags);
    printf("    field    : %u\n", buffer.field);
    printf("    timestamp: %ld.%06ld\n",
	   buffer.timestamp.tv_sec, buffer.timestamp.tv_usec);
    printf("    sequence : %u\n", buffer.field);
    usleep(100000); /* for proper printing */
  }

  /* Cleanup. */
  for (i = 0; i < reqbuf.count; i++)
    if (buffers[i].length) munmap (buffers[i].start, buffers[i].length);
  free(buffers);

  return 0;
}

int main(int argc, char *argv[])
{
  int video_dev;
  int ret;
  struct v4l2_input input;
  
  printf("\n\n   ***   test_v4l2   ***\n\n");
  printf("Basic test for V4L2 compatible PXA camera driver\n\n");
  usleep(100000); /* for proper printing */
  
  video_dev = open(VIDEO_DEV_NAME, O_RDWR);
  if (video_dev < 0) {
    printf("%s ", VIDEO_DEV_NAME);
    perror("open failed");
    exit(1);
  }

  printf("%s open succeeded.\n", VIDEO_DEV_NAME);

  /* device capabilitities */
  ret = query_capabilities(video_dev);
  if (ret)
    goto close_video_device;

  /* input features */
  ret = query_input(video_dev, &input);
  if (ret)
    goto close_video_device;
  
  /* supported video standards */
  ret = query_standards(video_dev, &input);
  if (ret)
    goto close_video_device;

  /* image formats */
  ret = query_formats(video_dev);
  if (ret)
    goto close_video_device;

  /* controls */
  ret = query_controls(video_dev);
  if (ret)
    goto close_video_device;

  /* crop */
  ret = query_crop(video_dev);
  if (ret)
    goto close_video_device;

  /* chip ID */
  ret = get_chip_ident(video_dev);
  if (ret)
    goto close_video_device;

  /* sensor register access */
  ret = reg_access(video_dev);
  if (ret)
    goto close_video_device;

  /* test buffer mapping with VGA resolution, YUV422 planarized format */
  ret = prepare_test_buffer_mapping(video_dev, 0, V4L2_PIX_FMT_YUV422P);
  if (ret)
    goto close_video_device;
  ret = test_buffer_mapping(video_dev);
  if (ret)
    goto close_video_device;

 close_video_device:
  usleep(100000); // for proper printing
  close(video_dev);
  exit(0);
}
